@@ -17,28 +17,26 @@
  * <http://www.gnu.org/licenses/>
  */
 
+import org.junit.Test
+import tech.libeufin.bank.DecimalNumber
+import tech.libeufin.bank.db.TransactionDAO.BankTransactionResult
+import tech.libeufin.bank.db.WithdrawalDAO.WithdrawalCreationResult
+import tech.libeufin.common.*
 import java.time.Instant
 import java.util.*
-import kotlin.test.*
-import org.junit.Test
-import org.postgresql.jdbc.PgConnection
-import tech.libeufin.bank.*
-import tech.libeufin.bank.TransactionDAO.*
-import tech.libeufin.bank.WithdrawalDAO.*
-import tech.libeufin.util.*
+import kotlin.test.assertEquals
 
 class AmountTest {
     // Test amount computation in database
     @Test
-    fun computationTest() = bankSetup { db ->  
-        val conn = db.dbPool.getConnection().unwrap(PgConnection::class.java)
-        conn.execSQLUpdate("UPDATE libeufin_bank.bank_accounts SET balance.val = 100000 WHERE internal_payto_uri = '$exchangePayto'")
+    fun computationTest() = bankSetup { db -> db.conn { conn ->
+        conn.execSQLUpdate("UPDATE libeufin_bank.bank_accounts SET balance.val = 100000 WHERE internal_payto_uri = '${customerPayto.canonical}'")
         val stmt = conn.prepareStatement("""
             UPDATE libeufin_bank.bank_accounts 
                 SET balance = (?, ?)::taler_amount
                     ,has_debt = ?
                     ,max_debt = (?, ?)::taler_amount
-            WHERE internal_payto_uri = '$merchantPayto'
+            WHERE internal_payto_uri = '${merchantPayto.canonical}'
         """)
         suspend fun routine(balance: TalerAmount, due: TalerAmount, hasBalanceDebt: Boolean, maxDebt: TalerAmount): Boolean {
             stmt.setLong(1, balance.value)
@@ -50,11 +48,12 @@ class AmountTest {
             // Check bank transaction
             stmt.executeUpdate()
             val txRes = db.transaction.create(
-                creditAccountPayto = exchangePayto,
+                creditAccountPayto = customerPayto,
                 debitAccountUsername = "merchant",
                 subject = "test",
                 amount = due,
                 timestamp = Instant.now(),
+                is2fa = false
             )
             val txBool = when (txRes) {
                 BankTransactionResult.BalanceInsufficient -> false
@@ -68,6 +67,7 @@ class AmountTest {
                 login = "merchant",
                 uuid = UUID.randomUUID(),
                 amount = due,
+                now = Instant.now()
             )
             val wBool = when (wRes) {
                 WithdrawalCreationResult.BalanceInsufficient -> false
@@ -122,29 +122,7 @@ class AmountTest {
             hasBalanceDebt = true,
             maxDebt = TalerAmount(0, 1, "KUDOS")
         ))
-
-
-    }
-
-    @Test
-    fun parse() {
-        assertEquals(TalerAmount("EUR:4"), TalerAmount(4L, 0, "EUR"))
-        assertEquals(TalerAmount("EUR:0.02"), TalerAmount(0L, 2000000, "EUR"))
-        assertEquals(TalerAmount("EUR:4.12"), TalerAmount(4L, 12000000, "EUR"))
-        assertEquals(TalerAmount("LOCAL:4444.1000"), TalerAmount(4444L, 10000000, "LOCAL"))
-        assertEquals(TalerAmount("EUR:${TalerAmount.MAX_VALUE}.99999999"), TalerAmount(TalerAmount.MAX_VALUE, 99999999, "EUR"))
-
-        assertException("Invalid amount format") {TalerAmount("")}
-        assertException("Invalid amount format") {TalerAmount("EUR")}
-        assertException("Invalid amount format") {TalerAmount("eur:12")}
-        assertException("Invalid amount format") {TalerAmount(" EUR:12")}
-        assertException("Invalid amount format") {TalerAmount("EUR:1.")}
-        assertException("Invalid amount format") {TalerAmount("EUR:.1")}
-        assertException("Invalid amount format") {TalerAmount("AZERTYUIOPQSD:12")}
-        assertException("Value specified in amount is too large") {TalerAmount("EUR:${Long.MAX_VALUE}")}
-        assertException("Invalid amount format") {TalerAmount("EUR:4.000000000")}
-        assertException("Invalid amount format") {TalerAmount("EUR:4.4a")}
-    }
+    }}
 
     @Test
     fun parseRoundTrip() {
@@ -262,6 +240,9 @@ class AmountTest {
                     }
                 }
             }
+            for (mode in listOf("zero", "up", "nearest")) {
+                assertEquals(TalerAmount("HUF:5"), apply(TalerAmount("HUF:5"), DecimalNumber("1"), DecimalNumber("1"), mode))
+            }
         }
     }
 
@@ -316,7 +297,7 @@ class AmountTest {
             for (mode in listOf("zero", "up", "nearest")) {
                 for (amount in listOf(10, 11, 12, 12, 14, 15, 16, 17, 18, 19)) {
                     for (tiny in listOf("0.01", "0.00000001", "5")) {
-                        for (ratio in listOf("0.341", "0.00000001")) {
+                        for (ratio in listOf("1", "0.341", "0.00000001")) {
                             val tiny = DecimalNumber(tiny)
                             val ratio = DecimalNumber(ratio)
                             val base = TalerAmount("EUR:$amount")
@@ -333,4 +314,19 @@ class AmountTest {
             }
         }
     }
+
+    @Test
+    fun apiError() = bankSetup { _ -> 
+        val base = obj {
+            "payto_uri" to "$exchangePayto?message=payout"
+        }
+
+        // Check OK
+        client.postA("/accounts/merchant/transactions") {
+            json(base) { "amount" to "KUDOS:0.3ABC" }
+        }.assertBadRequest(TalerErrorCode.BANK_BAD_FORMAT_AMOUNT)
+        client.postA("/accounts/merchant/transactions") {
+            json(base) { "amount" to "KUDOS:999999999999999999" }
+        }.assertBadRequest(TalerErrorCode.BANK_NUMBER_TOO_BIG)
+    }
 }
\ No newline at end of file
