@@ -1,6 +1,6 @@
 /*
  * This file is part of LibEuFin.
- * Copyright (C) 2023 Stanisci and Dold.
+ * Copyright (C) 2024 Taler Systems S.A.
 
  * LibEuFin is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Affero General Public License as
@@ -42,21 +42,20 @@ import io.ktor.client.plugins.*
 import io.ktor.client.request.*
 import io.ktor.client.statement.*
 import io.ktor.http.*
-import org.apache.commons.compress.archivers.zip.ZipFile
-import org.apache.commons.compress.utils.SeekableInMemoryByteChannel
-import org.slf4j.Logger
-import org.slf4j.LoggerFactory
+import io.ktor.utils.io.jvm.javaio.*
+import tech.libeufin.common.*
+import tech.libeufin.common.crypto.*
+import tech.libeufin.ebics.*
+import tech.libeufin.ebics.ebics_h005.Ebics3Request
 import tech.libeufin.nexus.*
-import tech.libeufin.util.*
-import tech.libeufin.util.ebics_h005.Ebics3Request
 import java.io.ByteArrayOutputStream
+import java.io.InputStream
+import java.io.SequenceInputStream
 import java.security.interfaces.RSAPrivateCrtKey
 import java.time.LocalDateTime
 import java.time.format.DateTimeFormatter
 import java.util.*
-import java.util.zip.DeflaterInputStream
-
-private val logger: Logger = LoggerFactory.getLogger("tech.libeufin.nexus.EbicsCommon")
+import kotlinx.coroutines.*
 
 /**
  * Available EBICS versions.
@@ -67,52 +66,13 @@ enum class EbicsVersion { two, three }
  * Which documents can be downloaded via EBICS.
  */
 enum class SupportedDocument {
-    /**
-     * Payment acknowledgement.
-     */
     PAIN_002,
-    /**
-     * From an HAC request.  Informs about any
-     * download/upload activity, including wrong
-     * documents.
-     */
     PAIN_002_LOGS,
-    /**
-     * Account statements.
-     */
     CAMT_053,
-    /**
-     * Account intraday reports.
-     */
     CAMT_052,
-    /**
-     * Account notifications.
-     */
     CAMT_054
 }
 
-
-/**
- * Unzips the ByteArray and runs the lambda over each entry.
- *
- * @param lambda function that gets the (fileName, fileContent) pair
- *        for each entry in the ZIP archive as input.
- */
-fun ByteArray.unzipForEach(lambda: (String, String) -> Unit) {
-    if (this.isEmpty()) {
-        tech.libeufin.nexus.logger.warn("Empty archive")
-        return
-    }
-    val mem = SeekableInMemoryByteChannel(this)
-    val zipFile = ZipFile(mem)
-    zipFile.getEntriesInPhysicalOrder().iterator().forEach {
-        lambda(
-            it.name, zipFile.getInputStream(it).readAllBytes().toString(Charsets.UTF_8)
-        )
-    }
-    zipFile.close()
-}
-
 /**
  * Decrypts and decompresses the business payload that was
  * transported within an EBICS message from the bank
@@ -128,53 +88,34 @@ fun decryptAndDecompressPayload(
     clientEncryptionKey: RSAPrivateCrtKey,
     encryptionInfo: DataEncryptionInfo,
     chunks: List<String>
-): ByteArray {
-    val buf = StringBuilder()
-    chunks.forEach { buf.append(it) }
-    val decoded = Base64.getDecoder().decode(buf.toString())
-    val er = CryptoUtil.EncryptionResult(
-        encryptionInfo.transactionKey,
-        encryptionInfo.bankPubDigest,
-        decoded
-    )
-    val dataCompr = CryptoUtil.decryptEbicsE002(
-        er,
-        clientEncryptionKey
-    )
-    return EbicsOrderUtil.decodeOrderData(dataCompr)
-}
+): InputStream =
+    SequenceInputStream(Collections.enumeration(chunks.map { it.toByteArray().inputStream() })) // Aggregate
+        .decodeBase64()
+        .run {
+            CryptoUtil.decryptEbicsE002(
+                encryptionInfo.transactionKey,
+                this,
+                clientEncryptionKey
+            )
+        }.inflate()
 
 /**
  * POSTs the EBICS message to the bank.
  *
  * @param URL where the bank serves EBICS requests.
- * @param msg EBICS message as raw string.
- * @return the raw bank response, if the request made it to the
- *         EBICS handler, or null otherwise.
+ * @param msg EBICS message as raw bytes.
+ * @return the raw bank response.
  */
-suspend fun HttpClient.postToBank(bankUrl: String, msg: String): String? {
-    logger.debug("POSTing EBICS to: $bankUrl")
-    val resp: HttpResponse = try {
-        this.post(urlString = bankUrl) {
-            expectSuccess = false // avoids exceptions on non-2xx statuses.
-            contentType(ContentType.Text.Xml)
-            setBody(msg)
-        }
+suspend fun HttpClient.postToBank(bankUrl: String, msg: ByteArray): InputStream {
+    logger.debug("POSTing EBICS to '$bankUrl'")
+    val res = post(urlString = bankUrl) {
+        contentType(ContentType.Text.Xml)
+        setBody(msg)
     }
-    catch (e: Exception) {
-        // hard error (network issue, invalid URL, ..)
-        tech.libeufin.nexus.logger.error("Could not POST to bank at: $bankUrl, detail: ${e.message}")
-        return null
+    if (res.status != HttpStatusCode.OK) {
+        throw Exception("Invalid response status: ${res.status}")
     }
-    // Bank was found, but the EBICS request wasn't served.
-    // Note: EBICS errors get _still_ 200 OK, so here the error
-    // _should_ not be related to EBICS.  404 for a wrong URL
-    // is one example.
-    if (resp.status != HttpStatusCode.OK) {
-        tech.libeufin.nexus.logger.error("Bank was found at $bankUrl, but EBICS wasn't served.  Response status: ${resp.status}, body: ${resp.bodyAsText()}")
-        return null
-    }
-    return resp.bodyAsText()
+    return res.bodyAsChannel().toInputStream()
 }
 
 /**
@@ -266,33 +207,50 @@ fun generateKeysPdf(
  *
  * @param clientKeys client keys, used to sign the request.
  * @param bankKeys bank keys, used to decrypt and validate the response.
- * @param xmlBody raw EBICS request in XML.
- * @param withEbics3 true in case the communication is EBICS 3, false otherwise.
- * @param tolerateEbicsReturnCode EBICS technical return code that may be accepted
- *                                instead of EBICS_OK.  That is the case of EBICS_DOWNLOAD_POSTPROCESS_DONE
- *                                along download receipt phases.
- * @param tolerateBankReturnCode Business return code that may be accepted instead of
- *                               EBICS_OK.  Typically, EBICS_NO_DOWNLOAD_DATA_AVAILABLE is tolerated
- *                               when asking for new incoming payments.
+ * @param xmlReq raw EBICS request in XML.
+ * @param isEbics3 true in case the communication is EBICS 3, false 
  * @return [EbicsResponseContent] or throws [EbicsSideException]
  */
 suspend fun postEbics(
     client: HttpClient,
     cfg: EbicsSetupConfig,
     bankKeys: BankPublicKeysFile,
-    xmlReq: String,
+    xmlReq: ByteArray,
     isEbics3: Boolean
 ): EbicsResponseContent {
-    val respXml = client.postToBank(cfg.hostBaseUrl, xmlReq)
-        ?: throw EbicsSideException(
+    val respXml = try {
+        client.postToBank(cfg.hostBaseUrl, xmlReq)
+    } catch (e: Exception) {
+        throw EbicsSideException(
             "POSTing to ${cfg.hostBaseUrl} failed",
-            sideEc = EbicsSideError.HTTP_POST_FAILED
+            sideEc = EbicsSideError.HTTP_POST_FAILED,
+            e
         )
-    return parseAndValidateEbicsResponse(
-        bankKeys,
-        respXml,
+    }
+
+    // Parses the bank response from the raw XML and verifies
+    // the bank signature.
+    val doc = try {
+        XMLUtil.parseIntoDom(respXml)
+    } catch (e: Exception) {
+        throw EbicsSideException(
+            "Bank response apparently invalid",
+            sideEc = EbicsSideError.BANK_RESPONSE_IS_INVALID
+        )
+    }
+    if (!XMLUtil.verifyEbicsDocument(
+        doc,
+        bankKeys.bank_authentication_public_key,
         isEbics3
-    )
+    )) {
+        throw EbicsSideException(
+            "Bank signature did not verify",
+            sideEc = EbicsSideError.BANK_SIGNATURE_DIDNT_VERIFY
+        )
+    }
+    if (isEbics3)
+        return ebics3toInternalRepr(doc)
+    return ebics25toInternalRepr(doc)
 }
 
 /**
@@ -306,8 +264,9 @@ private fun areCodesOk(ebicsResponseContent: EbicsResponseContent) =
             ebicsResponseContent.bankReturnCode == EbicsReturnCode.EBICS_OK
 
 /**
- * Collects all the steps of an EBICS download transaction.  Namely,
- * it conducts: init -> transfer -> receipt phases.
+ * Perform an EBICS download transaction.
+ * 
+ * It conducts init -> transfer -> processing -> receipt phases.
  *
  * @param client HTTP client for POSTing to the bank.
  * @param cfg configuration handle.
@@ -315,105 +274,115 @@ private fun areCodesOk(ebicsResponseContent: EbicsResponseContent) =
  * @param bankKeys bank EBICS public keys.
  * @param reqXml raw EBICS XML request of the init phase.
  * @param isEbics3 true for EBICS 3, false otherwise.
- * @param tolerateEmptyResult true if the EC EBICS_NO_DOWNLOAD_DATA_AVAILABLE
- *        should be tolerated as the bank-technical error, false otherwise.
- * @return the bank response as an uncompressed [ByteArray], or null if one
- *         error took place.  Definition of error: any EBICS- or bank-technical
- *         EC pairs where at least one is not EBICS_OK, or if tolerateEmptyResult
- *         is true, the bank-technical EC EBICS_NO_DOWNLOAD_DATA_AVAILABLE is allowed
- *         other than EBICS_OK.  If the request tolerates an empty download content,
- *         then the empty array is returned.  The function may throw [EbicsAdditionalErrors].
+ * @param processing processing lambda receiving EBICS files as a byte stream if the transaction was not empty.
+ * @return T if the transaction was successful. If the failure is at the EBICS 
+ *         level EbicsSideException is thrown else ités the exception of the processing lambda.
  */
-suspend fun doEbicsDownload(
+suspend fun ebicsDownload(
     client: HttpClient,
     cfg: EbicsSetupConfig,
     clientKeys: ClientPrivateKeysFile,
     bankKeys: BankPublicKeysFile,
-    reqXml: String,
+    reqXml: ByteArray,
     isEbics3: Boolean,
-    tolerateEmptyResult: Boolean = false
-): ByteArray? {
-    val initResp = postEbics(client, cfg, bankKeys, reqXml, isEbics3)
-    logger.debug("Download init phase done.  EBICS- and bank-technical codes are: ${initResp.technicalReturnCode}, ${initResp.bankReturnCode}")
-    if (initResp.technicalReturnCode != EbicsReturnCode.EBICS_OK) {
-        logger.error("Download init phase has EBICS-technical error: ${initResp.technicalReturnCode}")
-        return null
-    }
-    if (initResp.bankReturnCode == EbicsReturnCode.EBICS_NO_DOWNLOAD_DATA_AVAILABLE && tolerateEmptyResult) {
-        logger.info("Download content is empty")
-        return ByteArray(0)
-    }
-    if (initResp.bankReturnCode != EbicsReturnCode.EBICS_OK) {
-        logger.error("Download init phase has bank-technical error: ${initResp.bankReturnCode}")
-        return null
-    }
-    val tId = initResp.transactionID
-        ?: throw EbicsSideException(
-            "EBICS download init phase did not return a transaction ID, cannot do the transfer phase.",
-            sideEc = EbicsSideError.EBICS_UPLOAD_TRANSACTION_ID_MISSING
-        )
-    logger.debug("EBICS download transaction passed the init phase, got ID: $tId")
-    val howManySegments = initResp.numSegments
-    if (howManySegments == null) {
-        tech.libeufin.nexus.logger.error("Init response lacks the quantity of segments, failing.")
-        return null
-    }
-    val ebicsChunks = mutableListOf<String>()
-    // Getting the chunk(s)
-    val firstDataChunk = initResp.orderDataEncChunk
-        ?: throw EbicsSideException(
-            "OrderData element not found, despite non empty payload, failing.",
-            sideEc = EbicsSideError.ORDER_DATA_ELEMENT_NOT_FOUND
-        )
-    val dataEncryptionInfo = initResp.dataEncryptionInfo ?: run {
-        throw EbicsSideException(
-            "EncryptionInfo element not found, despite non empty payload, failing.",
-            sideEc = EbicsSideError.ENCRYPTION_INFO_ELEMENT_NOT_FOUND
-        )
-    }
-    ebicsChunks.add(firstDataChunk)
-    // proceed with the transfer phase.
-    for (x in 2 .. howManySegments) {
-        // request segment number x.
-        val transReq = if (isEbics3)
-            createEbics3DownloadTransferPhase(cfg, clientKeys, x, howManySegments, tId)
-        else createEbics25DownloadTransferPhase(cfg, clientKeys, x, howManySegments, tId)
-
-        val transResp = postEbics(client, cfg, bankKeys, transReq, isEbics3)
-        if (!areCodesOk(transResp)) {
+    processing: (InputStream) -> Unit
+) = coroutineScope {
+    val scope = this
+    // We need to run the logic in a non-cancelable context because we need to send 
+    // a receipt for each open download transaction, otherwise we'll be stuck in an 
+    // error loop until the pending transaction timeout.
+    // TODO find a way to cancel the pending transaction ?
+    withContext(NonCancellable) { 
+        val initResp = postEbics(client, cfg, bankKeys, reqXml, isEbics3) 
+        logger.debug("Download init phase done.  EBICS- and bank-technical codes are: ${initResp.technicalReturnCode}, ${initResp.bankReturnCode}")
+        if (initResp.technicalReturnCode != EbicsReturnCode.EBICS_OK) {
+            throw Exception("Download init phase has EBICS-technical error: ${initResp.technicalReturnCode}")
+        }
+        if (initResp.bankReturnCode == EbicsReturnCode.EBICS_NO_DOWNLOAD_DATA_AVAILABLE) {
+            logger.debug("Download content is empty")
+            return@withContext
+        } else if (initResp.bankReturnCode != EbicsReturnCode.EBICS_OK) {
+            throw Exception("Download init phase has bank-technical error: ${initResp.bankReturnCode}")
+        }
+        val tId = initResp.transactionID
+            ?: throw EbicsSideException(
+                "EBICS download init phase did not return a transaction ID, cannot do the transfer phase.",
+                sideEc = EbicsSideError.EBICS_UPLOAD_TRANSACTION_ID_MISSING
+            )
+        logger.debug("EBICS download transaction passed the init phase, got ID: $tId")
+        val howManySegments = initResp.numSegments
+        if (howManySegments == null) {
+            throw Exception("Init response lacks the quantity of segments, failing.")
+        }
+        val ebicsChunks = mutableListOf<String>()
+        // Getting the chunk(s)
+        val firstDataChunk = initResp.orderDataEncChunk
+            ?: throw EbicsSideException(
+                "OrderData element not found, despite non empty payload, failing.",
+                sideEc = EbicsSideError.ORDER_DATA_ELEMENT_NOT_FOUND
+            )
+        val dataEncryptionInfo = initResp.dataEncryptionInfo ?: run {
             throw EbicsSideException(
-                "EBICS transfer segment #$x failed.",
-                sideEc = EbicsSideError.TRANSFER_SEGMENT_FAILED
+                "EncryptionInfo element not found, despite non empty payload, failing.",
+                sideEc = EbicsSideError.ENCRYPTION_INFO_ELEMENT_NOT_FOUND
             )
         }
-        val chunk = transResp.orderDataEncChunk
-        if (chunk == null) {
-            tech.libeufin.nexus.logger.error("EBICS transfer phase lacks chunk #$x, failing.")
-            return null
-        }
-        ebicsChunks.add(chunk)
-    }
-    // all chunks gotten, shaping a meaningful response now.
-    val payloadBytes = decryptAndDecompressPayload(
-        clientKeys.encryption_private_key,
-        dataEncryptionInfo,
-        ebicsChunks
-    )
-    // payload reconstructed, receipt to the bank.
-    val receiptXml = if (isEbics3)
-        createEbics3DownloadReceiptPhase(cfg, clientKeys, tId)
-    else createEbics25DownloadReceiptPhase(cfg, clientKeys, tId)
+        ebicsChunks.add(firstDataChunk)
+        // proceed with the transfer phase.
+        for (x in 2 .. howManySegments) {
+            if (!scope.isActive) break
+            // request segment number x.
+            val transReq = if (isEbics3)
+                createEbics3DownloadTransferPhase(cfg, clientKeys, x, howManySegments, tId)
+            else createEbics25DownloadTransferPhase(cfg, clientKeys, x, howManySegments, tId)
 
-    // Sending the receipt to the bank.
-    postEbics(
-        client,
-        cfg,
-        bankKeys,
-        receiptXml,
-        isEbics3
-    )
-    // Receipt didn't throw, can now return the payload.
-    return payloadBytes
+            val transResp = postEbics(client, cfg, bankKeys, transReq, isEbics3)
+            if (!areCodesOk(transResp)) {
+                throw EbicsSideException(
+                    "EBICS transfer segment #$x failed.",
+                    sideEc = EbicsSideError.TRANSFER_SEGMENT_FAILED
+                )
+            }
+            val chunk = transResp.orderDataEncChunk
+            if (chunk == null) {
+                throw Exception("EBICS transfer phase lacks chunk #$x, failing.")
+            }
+            ebicsChunks.add(chunk)
+        }
+        suspend fun receipt(success: Boolean) {
+            val receiptXml = if (isEbics3)
+                createEbics3DownloadReceiptPhase(cfg, clientKeys, tId, success)
+            else createEbics25DownloadReceiptPhase(cfg, clientKeys, tId, success)
+
+            // Sending the receipt to the bank.
+            postEbics(
+                client,
+                cfg,
+                bankKeys,
+                receiptXml,
+                isEbics3
+            )
+        }
+        if (scope.isActive) {
+            // all chunks gotten, shaping a meaningful response now.
+            val payloadBytes = decryptAndDecompressPayload(
+                clientKeys.encryption_private_key,
+                dataEncryptionInfo,
+                ebicsChunks
+            )
+            // Process payload
+            val res = runCatching {
+                processing(payloadBytes)
+            }
+            receipt(res.isSuccess)
+
+            res.getOrThrow()
+        } else {
+            receipt(false)
+            throw CancellationException()
+        }
+    }
+    Unit
 }
 
 /**
@@ -445,45 +414,9 @@ enum class EbicsSideError {
  */
 class EbicsSideException(
     msg: String,
-    val sideEc: EbicsSideError
-) : Exception(msg)
-
-/**
- * Parses the bank response from the raw XML and verifies
- * the bank signature.
- *
- * @param bankKeys provides the bank auth pub, to verify the signature.
- * @param responseStr raw XML response from the bank
- * @param withEbics3 true if the communication is EBICS 3, false otherwise.
- * @return [EbicsResponseContent] or throw [EbicsSideException]
- */
-fun parseAndValidateEbicsResponse(
-    bankKeys: BankPublicKeysFile,
-    responseStr: String,
-    withEbics3: Boolean
-): EbicsResponseContent {
-    val responseDocument = try {
-        XMLUtil.parseStringIntoDom(responseStr)
-    } catch (e: Exception) {
-        throw EbicsSideException(
-            "Bank response apparently invalid",
-            sideEc = EbicsSideError.BANK_RESPONSE_IS_INVALID
-        )
-    }
-    if (!XMLUtil.verifyEbicsDocument(
-            responseDocument,
-            bankKeys.bank_authentication_public_key,
-            withEbics3
-    )) {
-        throw EbicsSideException(
-            "Bank signature did not verify",
-            sideEc = EbicsSideError.BANK_SIGNATURE_DIDNT_VERIFY
-        )
-    }
-    if (withEbics3)
-        return ebics3toInternalRepr(responseStr)
-    return ebics25toInternalRepr(responseStr)
-}
+    val sideEc: EbicsSideError,
+    cause: Exception? = null
+) : Exception(msg, cause)
 
 /**
  * Signs and the encrypts the data to send via EBICS.
@@ -530,9 +463,7 @@ fun prepareUploadPayload(
     val plainTransactionKey = encryptionResult.plainTransactionKey
         ?: throw Exception("Could not generate the transaction key, cannot encrypt the payload!")
     // Then only E002 symmetric (with ephemeral key) encrypt.
-    val compressedInnerPayload = DeflaterInputStream(
-        payload.inputStream()
-    ).use { it.readAllBytes() }
+    val compressedInnerPayload = payload.inputStream().deflate().readAllBytes()
     val encryptedPayload = CryptoUtil.encryptEbicsE002withTransactionKey(
         compressedInnerPayload,
         bankKeys.bank_encryption_public_key,
@@ -593,8 +524,8 @@ suspend fun doEbicsUpload(
     bankKeys: BankPublicKeysFile,
     orderService: Ebics3Request.OrderDetails.Service,
     payload: ByteArray,
-    extraLog: Boolean = false
-): EbicsResponseContent {
+): EbicsResponseContent = withContext(NonCancellable) {
+    // TODO use a lambda and pass the order detail there for atomicity ?
     val preparedPayload = prepareUploadPayload(cfg, clientKeys, bankKeys, payload, isEbics3 = true)
     val initXml = createEbics3RequestForUploadInitialization(
         cfg,
@@ -603,13 +534,12 @@ suspend fun doEbicsUpload(
         clientKeys,
         orderService
     )
-    if (extraLog) logger.debug(initXml)
     val initResp = postEbics( // may throw EbicsEarlyException
-            client,
-            cfg,
-            bankKeys,
-            initXml,
-            isEbics3 = true
+        client,
+        cfg,
+        bankKeys,
+        initXml,
+        isEbics3 = true
     )
     if (!areCodesOk(initResp)) throw EbicsUploadException(
         "EBICS upload init failed",
@@ -643,5 +573,5 @@ suspend fun doEbicsUpload(
         bankErrorCode = initResp.bankReturnCode
     )
     // EBICS- and bank-technical codes were both EBICS_OK, success!
-    return transferResp
+    transferResp
 }
\ No newline at end of file
